#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK

import logging

from ckan import plugins
from ckan.plugins import toolkit

from . import cli, helpers, routes
from .lib.doi import find_existing_doi, mint_multisearch_doi
from .lib.query import Query
from .lib.stats import DOWNLOAD_ACTION, record_stat
from .logic import action, auth

log = logging.getLogger(__name__)


class QueryDOIsPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.IBlueprint, inherit=True)
    plugins.implements(plugins.IConfigurer, inherit=True)
    plugins.implements(plugins.ITemplateHelpers)
    plugins.implements(plugins.IActions)
    plugins.implements(plugins.IAuthFunctions)
    plugins.implements(plugins.IClick)
    # if the versioned datastore downloader is available, we have a hook for it
    try:
        from ckanext.versioned_datastore.interfaces import IVersionedDatastoreDownloads

        plugins.implements(IVersionedDatastoreDownloads, inherit=True)
        versioned_datastore_available = True
    except ImportError:
        versioned_datastore_available = False

    # IBlueprint
    def get_blueprint(self):
        return routes.blueprints

    # IClick
    def get_commands(self):
        return cli.get_commands()

    # IAuthFunctions
    def get_auth_functions(self):
        return {
            'create_doi': auth.create_doi,
        }

    # IActions
    def get_actions(self):
        return {
            'create_doi': action.create_doi,
        }

    # IConfigurer
    def update_config(self, config):
        # add templates
        plugins.toolkit.add_template_directory(config, 'theme/templates')
        # add the resource groups
        plugins.toolkit.add_resource('theme/assets', 'ckanext-query-dois')

    # IVersionedDatastoreDownloads
    def download_after_init(self, request):
        try:
            query = Query.create_from_download_request(request)
            # mint the DOI on datacite if necessary
            mint_multisearch_doi(query)
        except toolkit.ValidationError:
            log.warning(
                'Could not create DOI for download, it contains private resources'
            )
        except:
            # if anything unexpected goes wrong we don't want to stop the download from
            # completing; just log the error and move on
            log.error('Failed to mint/retrieve DOI', exc_info=True)

    def download_modify_notifier_template_context(self, request, context):
        try:
            query = Query.create_from_download_request(request)
            # if a DOI can be created it should already have been created in
            # download_after_init
            doi = find_existing_doi(query)
            if doi:
                # update the context with the doi
                context['doi'] = doi.doi
        except:
            # if anything goes wrong we don't want to stop the download; just log the
            # error and move on
            log.error('Failed to retrieve DOI', exc_info=True)

        # always return the context
        return context

    def download_modify_manifest(self, manifest, request):
        try:
            query = Query.create_from_download_request(request)
            # if a DOI can be created it should already have been created in
            # download_after_init
            doi = find_existing_doi(query)
            if doi:
                # add the doi to the manifest
                manifest['query-doi'] = doi.doi
        except:
            # if anything goes wrong we don't want to stop the download from completing;
            # just log the error and move on
            log.error('Failed to retrieve DOI', exc_info=True)

        # always return the manifest
        return manifest

    def download_after_run(self, request):
        try:
            query = Query.create_from_download_request(request)
            # if a DOI can be created it should already have been created in
            # download_modify_manifest
            doi = find_existing_doi(query)
            if doi and request.state == 'complete':
                # record a download stat against the DOI
                record_stat(doi, DOWNLOAD_ACTION, identifier=request.id)
        except:
            # just log the error and move on
            log.error('Failed to retrieve DOI and/or create stats', exc_info=True)

    # ITemplateHelpers
    def get_helpers(self):
        return {
            'render_filter_value': helpers.render_filter_value,
            'get_most_recent_dois': helpers.get_most_recent_dois,
            'get_time_ago_description': helpers.get_time_ago_description,
            'get_landing_page_url': helpers.get_landing_page_url,
            'create_citation_text': helpers.create_citation_text,
            'create_multisearch_citation_text': helpers.create_multisearch_citation_text,
            'pretty_print_query': helpers.pretty_print_query,
            'get_doi_count': helpers.get_doi_count,
            'versioned_datastore_available': self.versioned_datastore_available,
        }



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK


from ckan.model import DomainObject, meta
from sqlalchemy import BigInteger, Column, DateTime, Table, UnicodeText
from sqlalchemy.dialects.postgresql import JSONB

query_doi_table = Table(
    'query_doi',
    meta.metadata,
    Column('id', BigInteger, primary_key=True),
    # the full doi (prefix/suffix)
    Column('doi', UnicodeText, nullable=False, index=True, unique=True),
    # json column representing the resources in this query and their rounded versions, it is a
    # straight map from resource_id: version
    Column('resources_and_versions', JSONB, nullable=False),
    # the timestamp when the doi was created
    Column('timestamp', DateTime, nullable=False),
    # the query dict that produces the data for this doi
    Column('query', JSONB, nullable=False),
    # the hash for the query that produces the data for this doi - this is used in conjunction with
    # the version to check if the query has been run before
    Column('query_hash', UnicodeText, nullable=False, index=True),
    # the version initially requested by the user
    Column('requested_version', BigInteger, nullable=True),
    # record count at time of minting
    Column('count', BigInteger, nullable=False),
    # record the query version
    Column('query_version', UnicodeText, nullable=True),
    # record the resource counts
    Column('resource_counts', JSONB, nullable=True),
)


query_doi_stat_table = Table(
    'query_doi_stat',
    meta.metadata,
    Column('id', BigInteger, primary_key=True),
    # the doi this stat relates to
    Column('doi', UnicodeText, nullable=False, index=True),
    # record the action that produced this stat entry (for example, search or download)
    Column('action', UnicodeText),
    # the domain from the email address of the user using the doi
    Column('domain', UnicodeText),
    # the encrypted identifier from the email address of the user using the doi
    Column('identifier', UnicodeText),
    # timestamp of the stat
    Column('timestamp', DateTime, nullable=False),
)


class QueryDOI(DomainObject):
    """
    Object for holding query DOIs.
    """

    def get_resource_ids(self):
        return list(self.resources_and_versions.keys())

    def get_rounded_versions(self):
        return list(self.resources_and_versions.values())

    @staticmethod
    def on_resource(resource_id):
        """
        A convenience method to filter by a specific resource id.

        :param resource_id: the resource id
        :returns: an sqlalchemy boolean expression
        """
        return QueryDOI.resources_and_versions.has_key(resource_id)


class QueryDOIStat(DomainObject):
    """
    Object for holding query DOIs stats.
    """

    def to_dict(self):
        """
        Returns the object as a dict for the stats API response.

        :returns: a dict
        """
        return {
            'id': self.id,
            'doi': self.doi,
            'action': self.action,
            'domain': self.domain,
            'identifier': self.identifier,
            'timestamp': str(self.timestamp),
        }


meta.mapper(QueryDOI, query_doi_table)
meta.mapper(QueryDOIStat, query_doi_stat_table)



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK

from ckan.plugins import toolkit

from ckanext.query_dois.lib.doi import mint_multisearch_doi
from ckanext.query_dois.lib.emails import send_saved_search_email
from ckanext.query_dois.lib.query import Query
from ckanext.query_dois.lib.stats import SAVE_ACTION, record_stat
from ckanext.query_dois.logic import schema as schema_lib


def create_doi(context, data_dict):
    """
    Creates a DOI using the given parameters and returns it.

    :param email_address: the email address of the DOI requester
    :type email_address: string
    :param query: the query to associate with the DOI
    :type query: dict
    :param query_version: the query schema version for the query
    :type query_version: string
    :param version: the version to search the data at
    :type version: int, number of milliseconds (not seconds!) since UNIX epoch
    :param resource_ids: the resource ids to search
    :type resource_ids: list of strings
    :param doi: the doi that was created (prefix/suffix)
    :type doi: string
    :param is_new: whether the doi was newly created or whether an existing DOI for the
        query parameters already existed
    :type is_new: bool
    :param email_sent: whether the email was sent successfully or not
    :type email_sent: bool
    :rtype: dict
    """
    # validate the data dict first
    schema = context.get('schema', schema_lib.create_doi())
    data_dict, errors = toolkit.navl_validate(data_dict, schema, context)
    if errors:
        raise toolkit.ValidationError(errors)

    email_address = data_dict['email_address']
    query = Query.create(
        data_dict['resource_ids'],
        data_dict.get('version'),
        data_dict.get('query'),
        data_dict.get('query_version'),
    )

    # create a new DOI or retrieve an existing one
    created, doi = mint_multisearch_doi(query)
    # record a stat for this action
    record_stat(doi, SAVE_ACTION, email_address)
    # send the email to the requesting user
    email_sent = send_saved_search_email(email_address, doi)

    return {'is_new': created, 'doi': doi.doi, 'email_sent': email_sent}



# -*- coding: utf-8 -*-

from __future__ import with_statement

import os
from logging.config import fileConfig

from alembic import context
from sqlalchemy import engine_from_config, pool

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = None

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.

name = os.path.basename(os.path.dirname(__file__))


def run_migrations_offline():
    """
    Run migrations in 'offline' mode.

    This configures the context with just a URL and not an Engine, though an Engine is
    acceptable here as well.  By skipping the Engine creation we don't even need a DBAPI
    to be available.

    Calls to context.execute() here emit the given string to the script output.
    """

    url = config.get_main_option('sqlalchemy.url')
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        version_table='{}_alembic_version'.format(name),
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """
    Run migrations in 'online' mode.

    In this scenario we need to create an Engine and associate a connection with the
    context.
    """
    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix='sqlalchemy.',
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            version_table='{}_alembic_version'.format(name),
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK

from ckan.plugins import toolkit

from ckanext.datastore.logic.schema import json_validator

# grab all the validator functions upfront
boolean_validator = toolkit.get_validator('boolean_validator')
ignore_missing = toolkit.get_validator('ignore_missing')
int_validator = toolkit.get_validator('int_validator')
email_validator = toolkit.get_validator('email_validator')


def list_of_strings(delimiter=','):
    """
    Creates a converter/validator function which when given a value return a list or
    raises an error if a list can't be created from the value. If the value passed in is
    a list already it is returned with no modifications, if it's a string then the
    delimiter is used to split the string and the result is returned. If the value is
    neither a list or a string then an error is raised.

    :param delimiter: the string to delimit the value on, if it's a string. Defaults to
        a comma
    :returns: a list
    """

    def validator(value):
        if isinstance(value, list):
            return value
        if isinstance(value, str):
            return value.split(delimiter)
        raise toolkit.Invalid('Invalid list of strings')

    return validator


def create_doi():
    return {
        'resource_ids': [list_of_strings()],
        'email_address': [email_validator],
        'query': [ignore_missing, json_validator],
        'query_version': [ignore_missing, str],
        'version': [ignore_missing, int_validator],
    }



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK

import itertools
import time
from dataclasses import dataclass
from functools import cached_property, partial
from typing import Dict, List, Optional

from ckan import model
from ckan.plugins import toolkit
from sqlalchemy import false


def find_invalid_resources(resource_ids: List[str]) -> List[str]:
    """
    Given a list of resource IDs, return a list of resource IDs which are invalid.
    Resources are invalid if they are any of the following:

        - not datastore active resources (checked with vds_resource_check)
        - not active
        - not in an active package
        - not in a public package

    :param resource_ids: the resource IDs to check
    :returns: a list of resource IDs which failed the tests
    """
    invalid_resource_ids = set()

    # cache this action (with context) so that we don't have to retrieve it over and
    # over again
    is_datastore_resource = partial(toolkit.get_action('vds_resource_check'), {})

    # retrieve all resource ids passed to this function that are also active, in an
    # active package and in a public package
    query = (
        model.Session.query(model.Resource)
        .join(model.Package)
        .filter(model.Resource.id.in_(list(resource_ids)))
        .filter(model.Resource.state == 'active')
        .filter(model.Package.state == 'active')
        .filter(model.Package.private == false())
        .with_entities(model.Resource.id)
    )
    # go through each resource ID we found and check if they are datastore resources
    for row in query:
        if not is_datastore_resource(dict(resource_id=row.id)):
            invalid_resource_ids.add(row.id)

    return sorted(invalid_resource_ids)


@dataclass(frozen=True)
class Query:
    """
    Class representing a query against the versioned datastore.
    """

    resource_ids: List[str]
    version: int
    query: dict
    query_version: str

    @cached_property
    def query_hash(self) -> str:
        """
        :returns: a unique hash made from the query and query version
        """
        return toolkit.get_action('vds_multi_hash')(
            {}, {'query': self.query, 'query_version': self.query_version}
        )

    @cached_property
    def authors(self) -> List[str]:
        """
        Given some resource ids, return a list of unique authors from the packages
        associated with them.

        :returns: a list of authors
        """
        query = (
            model.Session.query(model.Resource)
            .join(model.Package)
            .filter(model.Resource.id.in_(self.resource_ids))
            .with_entities(model.Package.author)
        )
        return list(set(itertools.chain.from_iterable(query)))

    @cached_property
    def resources_and_versions(self) -> Dict[str, int]:
        """
        Returns a dict containing the resource IDs as keys and their rounded versions as
        values. The rounded versions are acquired via the vds_version_round action.

        :returns: a dict of resource IDs to rounded versions
        """
        action = toolkit.get_action('vds_version_round')
        return {
            resource_id: action(
                {}, {'resource_id': resource_id, 'version': self.version}
            )
            for resource_id in sorted(self.resource_ids)
        }

    @cached_property
    def counts(self) -> Dict[str, int]:
        """
        Returns a dict containing the resource IDs as keys and the number of records
        which match this query in the resource as the values.

        :returns: a dict of resource ids to counts
        """
        data_dict = {
            'query': self.query,
            'query_version': self.query_version,
            'resource_ids': self.resource_ids,
            'version': self.version,
        }
        return toolkit.get_action('vds_multi_count')({}, data_dict)['counts']

    @cached_property
    def count(self) -> int:
        """
        The total number of records matching this query.

        :returns: an integer
        """
        return sum(self.counts.values())

    @classmethod
    def create(
        cls,
        resource_ids: List[str],
        version: Optional[int] = None,
        query: Optional[dict] = None,
        query_version: Optional[str] = None,
    ) -> 'Query':
        """
        Creates a Query object using the given parameters. The resource_ids are the only
        required parameters, everything else is optional and will be defaulted to
        sensible values if needed.

        :param resource_ids: the resource IDs
        :param version: the version to query at (if missing, defaults to now)
        :param query: the query to run (if missing, defaults to any empty query)
        :param query_version: the version of the query (if missing, defaults to the
            latest query schema version)
        :returns: a Query object
        """
        invalid_resource_ids = find_invalid_resources(resource_ids)
        if invalid_resource_ids:
            # not all of them were public/active
            raise toolkit.ValidationError(
                f'Some of the resources requested are private or not active, DOIs can '
                f'only be created using public, active resources. Invalid resources: '
                f'{", ".format(invalid_resource_ids)}'
            )

        # sort them to ensure comparisons work consistently
        resource_ids = sorted(resource_ids)
        # default the version to now if not provided
        version = version if version is not None else int(time.time() * 1000)
        query = query or {}
        query_version = query_version or toolkit.get_action('vds_schema_latest')({}, {})

        return cls(resource_ids, version, query, query_version)

    @classmethod
    def create_from_download_request(cls, download_request):
        """
        Given a download request from the vds, turn it into our representation of a
        query.

        :param download_request: a DownloadRequest object from vds
        """
        return Query.create(
            download_request.core_record.resource_ids_and_versions,
            download_request.core_record.get_version(),
            download_request.core_record.query,
            download_request.core_record.query_version,
        )



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK
import json
from datetime import datetime
from typing import Optional

from ckan import model
from ckan.lib.helpers import link_to
from ckan.plugins import toolkit
from sqlalchemy import or_
from sqlalchemy.orm import Query

from ckanext.query_dois.model import QueryDOI


def render_filter_value(field, filter_value):
    """
    Renders the given filter value for the given field. This should be called for each
    filter value rather than by passing a list or filter values.

    :param field: the field name
    :param filter_value: the filter value for the field
    :returns: the value to display
    """
    if field == '__geo__':
        return json.loads(filter_value)['type']
    else:
        return filter_value


def _make_all_resource_query(package_id: str) -> Optional[Query]:
    """
    Creates an SQL Alchemy query that can get all the QueryDOI entities associated with
    the resources of the given package.

    :param package_id:
    :returns: None if the package doesn't exist or if the package has no resources,
        otherwise, returns an SQL Alchemy Query object
    """
    try:
        package = toolkit.get_action('package_show')({}, {'id': package_id})
    except toolkit.ObjectNotFound:
        return None
    ors = [QueryDOI.on_resource(resource['id']) for resource in package['resources']]
    if not ors:
        return None
    return model.Session.query(QueryDOI).filter(or_(*ors))


def get_most_recent_dois(package_id, number):
    """
    Retrieve the most recent DOIs that have been minted on queries against the resources
    in the given package.

    :param package_id: the package's ID
    :param number: the number of DOIs to return
    :returns: a list of QueryDOI objects
    """
    query = _make_all_resource_query(package_id)
    if query is None:
        return []
    return list(query.order_by(QueryDOI.id.desc()).limit(number))


def get_doi_count(package_id: str) -> int:
    """
    Return the total count of DOIs created against the resources within the given
    package.

    :param package_id: the ID of the package
    :returns: a number
    """
    query = _make_all_resource_query(package_id)
    return 0 if query is None else query.count()


# a tuple describing various ways of informing the user something happened a certain number of time
# units ago
time_resolutions = (
    (60, 'seceond', 1),
    (60 * 60, 'minute', 60),
    (60 * 60 * 24, 'hour', 60 * 60),
    (60 * 60 * 24 * 7, 'day', 60 * 60 * 24),
    (60 * 60 * 24 * 28, 'week', 60 * 60 * 24 * 7),
    (60 * 60 * 24 * 365, 'month', 60 * 60 * 24 * 28),
    (float('inf'), 'year', 60 * 60 * 24 * 365),
)


def get_time_ago_description(query_doi):
    """
    Given a QueryDOI object, return a short description of how long ago it was minted.
    The resolutions are described above in the time_resolutions tuple.

    :param query_doi: the QueryDOI object
    :returns: a unicode string describing how long ago the DOI was minted
    """
    seconds = (datetime.now() - query_doi.timestamp).total_seconds()
    for limit, unit, divisor in time_resolutions:
        if seconds < limit:
            value = int(seconds / divisor)
            plural = 's' if value > 1 else ''
            return f'{value} {unit}{plural} ago'


def get_landing_page_url(query_doi):
    """
    Given a QueryDOI object, return the landing URL for it.

    :param query_doi: a QueryDOI object
    :returns: the landing page URL
    """
    data_centre, identifier = query_doi.doi.split('/')
    return toolkit.url_for(
        'query_doi.landing_page', data_centre=data_centre, identifier=identifier
    )


def create_citation_text(
    query_doi,
    creation_timestamp,
    resource_name,
    package_title,
    package_doi=None,
    publisher=None,
    html=False,
):
    """
    Creates the citation text for the given query doi and the given additional related
    arguments.

    :param query_doi: the query's DOI, this should just be the prefix/suffix, e.g.
        10.xxxx/xxxxxx, not the full URL
    :param creation_timestamp: a datetime object representing the exact moment the DOI
        was created
    :param resource_name: the name of the resource the DOI references
    :param package_title: the title of the package the resource the DOI references is in
    :param package_doi: the DOI of the package, if there is one (defaults to None)
    :param publisher: the publisher to use in the citation (defaults to None in which
        case the ckanext.query_dois.publisher config value is used
    :param html: whether to include a tags around URLs in the returned string. Defaults
        to False which does not add a tags and therefore the returned string is just
        pure text
    :returns: a citation string for the given query DOI and associated data
    """
    # default the publisher's value if needed
    if publisher is None:
        publisher = toolkit.config.get('ckanext.query_dois.publisher')

    # this is the citation's base form. This form is derived from the recommended RDA citation
    # format for evolving data when citing with a query. For more information see:
    # https://github.com/NaturalHistoryMuseum/ckanext-query-dois/issues/2
    citation_text = (
        '{publisher} ({year}). Data Portal Query on "{resource_name}" created at '
        '{creation_datetime} PID {query_doi}. Subset of "{dataset_name}" (dataset)'
    )

    # these are the parameters which will be used on the above string
    params = {
        'publisher': publisher,
        'year': creation_timestamp.year,
        'resource_name': resource_name,
        'creation_datetime': creation_timestamp,
        'query_doi': 'https://doi.org/{}'.format(query_doi),
        'dataset_name': package_title,
    }

    # if we have a DOI for the package, include it
    if package_doi is not None:
        citation_text += ' PID {dataset_doi}.'
        params['dataset_doi'] = f'https://doi.org/{package_doi}'

    if html:
        # there are currently two fields which should be converted to a tags
        for field in ('query_doi', 'dataset_doi'):
            doi_url = params.get(field, None)
            # the dataset_doi field is optional, hence the if here
            if doi_url is not None:
                params[field] = link_to(label=doi_url, url=doi_url, target='_blank')

    return citation_text.format(**params)


def create_multisearch_citation_text(query_doi, html=False):
    """
    Creates the citation text for a multisearch based DOI and returns it, either as a
    raw string or as a piece of html, if requested.

    :param query_doi: a query doi object
    :param html: whether to return a basic string or a piece of html
    :returns: the citation text
    """
    publisher = toolkit.config.get('ckanext.query_dois.publisher')

    # this is the citation's base form. This form is derived from the recommended RDA citation
    # format for evolving data when citing with a query. For more information see:
    # https://github.com/NaturalHistoryMuseum/ckanext-query-dois/issues/2
    citation_text = (
        '{publisher} ({year}). Data Portal query on {resource_count} resources '
        'created at {creation_datetime} PID {query_doi}'
    )

    # these are the parameters which will be used on the above string
    params = {
        'publisher': publisher,
        'year': query_doi.timestamp.year,
        'resource_count': len(query_doi.get_resource_ids()),
        'creation_datetime': query_doi.timestamp,
    }

    doi_url = f'https://doi.org/{query_doi.doi}'
    if html:
        params['query_doi'] = link_to(label=doi_url, url=doi_url, target='_blank')
    else:
        params['query_doi'] = doi_url

    return citation_text.format(**params)


def pretty_print_query(query):
    """
    Does what you'd expect really.

    :param query: a query dict
    :returns: a string of pretty json
    """
    return json.dumps(query, sort_keys=True, indent=2)
